@@ -3,7 +3,7 @@ Interfaces to remote engines, i.e. filesystems, SQL & NoSQL databases, etc..
 """
 import pathlib
 import typing
-from io import TextIOWrapper
+import io
 
 
 class FileSystem(object):
@@ -221,19 +221,9 @@ class S3Connection(FileSystem):
         if "r" in mode:
             resp = self._client.get_object(Bucket=self._bucket, Key=path)
             stream = resp["Body"]
-            # make 'stream' more file-like
-            stream.readable = lambda: True
-            stream.writable = lambda: False
-            stream.seekable = lambda: False
-            if not hasattr(stream, "closed"):
-                stream.closed = False
             if mode == "r":
-                stream = TextIOWrapper(stream, 'utf-8')
-            if not hasattr(stream, "flush"):
-                stream.flush = lambda: None
-            stream.__enter__ = lambda *a, **k: stream
-            stream.__exit__ = lambda *a, **k: stream.close()
-            return stream
+                stream = io.TextIOWrapper(stream, 'utf-8')
+            return GenericStreamWrapper(stream, writable=False)
         if self._read_only:
             raise PermissionError("this data source is read-only")
         if "w" in mode:
@@ -270,3 +260,57 @@ class S3Connection(FileSystem):
 
     def rm(self, path, **kwargs):
         raise NotImplementedError()
+
+
+class GenericStreamWrapper(io.RawIOBase):
+    def __init__(self, stream, readable=None, writable=None):
+        self._target = stream
+        self._readable = readable if readable is not None else True if not hasattr(stream, "readable") else stream.readable()
+        self._writable = writable if writable is not None else False if not hasattr(stream, "writable") else stream.writable()
+        self._seekable = False if not hasattr(stream, "seekable") else stream.seekable()
+        self._closed = False
+
+    def readable(self) -> bool:
+        return self._readable
+
+    def writable(self) -> bool:
+        return self._writable
+
+    def seekable(self) -> bool:
+        return self._seekable
+
+    @property
+    def closed(self):
+        if not hasattr(self._target, "closed"):
+            return self._closed
+        return self._target.closed
+
+    def close(self):
+        self._closed = True
+        if not hasattr(self._target, "close"):
+            return
+        self._target.close()
+
+    def read(self, __size: int=None):
+        return self._target.read(__size)
+
+    def write(self, __b):
+        if not hasattr(self._target, "write"):
+            raise NotImplementedError()
+        return self._target.write(__b)
+
+    def seek(self, __offset: int, __whence: int=None) -> int:
+        if not hasattr(self._target, "seek"):
+            raise NotImplementedError()
+        return self._target.seek(__offset, __whence)
+
+    def flush(self) -> None:
+        if not hasattr(self._target, "flush"):
+            return
+        self._target.flush()
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        self.close()
